import numpy as np
from numpy import ma, random

from scipy.interpolate import RectBivariateSpline, interp1d

# ============================================================================
# I need to improve this, and include the places where the
#   flags are masked, i.e. only eliminate where the flags
#   could guarantee it was false.

def combined_flag(flags, criteria=None):
    """ Returns the combined flag considering all the criteria

        Input: flags

        Collects all flags in the criteria, and for each measurements, it
          return the maximum flag value among the different criteria.

        If criteria is not defined, considers all the flags,
          i.e. flags.keys()
    """
    assert hasattr(flags, 'keys')

    if criteria is None:
        criteria = flags.keys()

    output = np.asanyarray(flags[criteria[0]])
    for c in criteria[1:]:
        assert len(flags[c]) == len(output)
        output = np.max([output, flags[c]], axis=0)

    return output


def make_qc_index(flags, criteria, type="anytrue"):
    ind = flags[criteria[0]].copy()
    if type == "anytrue":
        for c in criteria:
            ind[(ind == True) | (flags[c] == True)] = True
        #ind[np.nonzero((ind == True) | (np.array(flags[c]) == True))[0]] = True
    elif type == "alltrue":
        for c in criteria:
            ind[(ind == True) | (flags[c] == True)] = True
    for c in criteria:
        ind[(ind == False) | (flags[c] == False)] = False
        #ind[np.nonzero((ind == False) | (np.array(flags[c]) == False))[0]] = False
    return ind



""" Apply Quality Control of CTD profiles
"""

import pkg_resources
from datetime import datetime
from os.path import basename, expanduser
import re
import json
import logging

import numpy as np
from numpy import ma

from seabird import cnv, CNVError
from seabird.utils import basic_logger
logging.basicConfig(level=logging.DEBUG)

from cotede.qctests import *
from cotede.misc import combined_flag
from cotede.utils import get_depth, woa_profile


class ProfileQC(object):
    """ Quality Control of a CTD profile
    """
    def __init__(self, input, cfg=None, saveauxiliary=True, verbose=True,
            attributes=None, logger=None):
        """
            Input: dictionary with data.
                - pressure[\d]:
                - temperature[\d]:
                - salinity[\d]:

            cfg: config file with thresholds

            =======================
            - Must have a log system
            - Probably accept incomplete cfg. If some threshold is
                not defined, take the default value.
            - Is the best return another dictionary?
        """
        self.logger = logger or logging.getLogger(__name__)

        self.name = input.filename
        self.verbose = verbose

        if attributes is None:
            assert (hasattr(input, 'attributes'))
        assert (hasattr(input, 'keys')) and (len(input.keys()) > 0)

        self.load_cfg(cfg)

        self.input = input
        if attributes is None:
            self.attributes = input.attributes
        else:
            self.attributes = attributes
        self.flags = {}
        self.saveauxiliary = saveauxiliary
        if saveauxiliary:
            #self.auxiliary = {}
            self.build_auxiliary()

        # I should use common or main, but must be consistent
        #   between defaults and flags.keys()
        # Think about it
        self.evaluate_common(self.cfg)

        for v in self.input.keys():
            for c in self.cfg.keys():
                if re.match("%s\d?$" % c, v):
                    if verbose is True:
                        self.logger.debug(" %s - evaluating: %s, as type: %s" %
                                (self.name, v, c))
                    self.evaluate(v, self.cfg[c])
                    break

    @property
    def data(self):
        return self.input.data

    def keys(self):
        """ Return the available keys in self.data
        """
        return self.input.keys()

    def __getitem__(self, key):
        """ Return the key array from self.data
        """
        return self.input[key]

    def load_cfg(self, cfg):
        """ Load the QC configurations

            The possible inputs are:
                - None: Will use the CoTeDe's default configuration

                - Preset config name [string]: A string with the name of
                    pre-set rules, like 'cotede', 'egoos' or 'gtspp'.

                - User configs [dict]: a dictionary composed by the variables
                    to be evaluated as keys, and inside it another dictionary
                    with the tests to perform as keys. example
                    {'main':{
                        'valid_datetime': None,
                        },
                    'temperature':{
                        'global_range':{
                            'minval': -2.5,
                            'maxval': 45,
                            },
                        },
                    }
        """
        # A given manual configuration has priority
        if type(cfg) is dict:
            self.cfg = cfg
            self.logger.debug("%s - User's QC cfg." % self.name)
            return

        # Need to safe_eval before allow to load rules from .cotederc
        if cfg is None:
            cfg = 'cotede'

        # If it's a name of a config, try to get from CoTeDe's package
        try:
            self.cfg = json.loads(pkg_resources.resource_string(__name__,
                "qc_cfg/%s" % cfg))
            if self.verbose is True:
                self.logger.debug("%s - QC cfg: %s" % (self.name, cfg))
        # If can't find inside cotede, try to load from users directory
        except:
            self.cfg = json.loads(expanduser('~/.cotederc/%s' % cfg))
            if self.verbose is True:
                self.logger.debug("%s - QC cfg: ~/.cotederc/%s" %
                        (self.name, cfg))

    def evaluate_common(self, cfg):
        if 'main' not in self.cfg.keys():
            self.logger.warn("ATTENTION, there is no main setup in the QC cfg")
            return

        self.flags['common'] = {}

        if 'valid_datetime' in self.cfg['main']:
            if 'datetime' in self.attributes.keys() and \
                    type(self.attributes['datetime']) == datetime:
                f = 1
            else:
                f = 3
            self.flags['common']['valid_datetime'] = f

        if 'datetime_range' in self.cfg['main']:
            if 'datetime' in self.attributes.keys() and \
                    (self.attributes['datetime'] >=
                            self.cfg['main']['datetime_range']['minval']) and \
                    (self.attributes['datetime'] <=
                            self.cfg['main']['datetime_range']['maxval']):
                f = 1
            else:
                f = 3
            self.flags['common']['datetime_range'] = f

        if 'at_sea' in self.cfg['main']:
            self.flags['common']['at_sea'] = location_at_sea(
                    self.input,
                    self.cfg['main']['at_sea'])

        if self.saveauxiliary:
            self.auxiliary['common'] = {}
            self.auxiliary['common']['descentPrate'] = \
                    descentPrate(self['timeS'], self['pressure'])

    def evaluate(self, v, cfg):

        self.flags[v] = {}

        if self.saveauxiliary:
            if v not in self.auxiliary.keys():
                self.auxiliary[v] = {}

        if 'global_range' in cfg:
            self.flags[v]['global_range'] = np.zeros(self.input[v].shape,
                    dtype='i1')
            # Flag as 9 any masked input value
            self.flags[v]['global_range'][ma.getmaskarray(self.input[v])] = 9
            ind = (self.input[v] >= cfg['global_range']['minval']) & \
                    (self.input[v] <= cfg['global_range']['maxval'])
            self.flags[v]['global_range'][np.nonzero(ind)] = 1
            ind = (self.input[v] < cfg['global_range']['minval']) | \
                    (self.input[v] > cfg['global_range']['maxval'])
            self.flags[v]['global_range'][np.nonzero(ind)] = 4

        if 'profile_envelope' in cfg:
            # Probably not the best way to do this, but works for now.
            self.flags[v]['profile_envelope'] = np.zeros(self.input[v].shape,
                    dtype='i1')
            for layer in cfg['profile_envelope']:
                ind = np.nonzero(
                        eval("(%s %s) & (%s %s)" %
                            ("self.input['pressure']", layer[0],
                                "self.input['pressure']", layer[1]))
                            )[0]
                f = eval("(%s > %s) & (%s < %s)" %
                        ("self.input[v][ind]", layer[2],
                        "self.input[v][ind]", layer[3]))
                self.flags[v]['profile_envelope'][ind[f == True]] = 1
                self.flags[v]['profile_envelope'][ind[f == False]] = 4

        if 'gradient' in cfg:
            threshold = cfg['gradient']
            g = gradient(self.input[v])

            if self.saveauxiliary:
                self.auxiliary[v]['gradient'] = g

            flag = np.zeros(g.shape, dtype='i1')
            flag[self.input[v].mask == True] = 9
            flag[np.nonzero(g > threshold)] = 4
            flag[np.nonzero(g <= threshold)] = 1
            self.flags[v]['gradient'] = flag

        if 'gradient_depthconditional' in cfg:
            cfg_tmp = cfg['gradient_depthconditional']
            g = gradient(self.input[v])
            flag = np.zeros(g.shape, dtype='i1')
            # Flag as 9 any masked input value
            flag[ma.getmaskarray(self.input[v])] = 9
            # ---- Shallow zone -----------------
            threshold = cfg_tmp['shallow_max']
            flag[np.nonzero( \
                    (self['pressure'] <= cfg_tmp['pressure_threshold']) & \
                    (g > threshold))] \
                    = 4
            flag[np.nonzero( \
                    (self['pressure'] <= cfg_tmp['pressure_threshold']) & \
                    (g <= threshold))] \
                    = 1
            # ---- Deep zone --------------------
            threshold = cfg_tmp['deep_max']
            flag[np.nonzero( \
                    (self['pressure'] > cfg_tmp['pressure_threshold']) & \
                    (g > threshold))] \
                    = 4
            flag[np.nonzero( \
                    (self['pressure'] > cfg_tmp['pressure_threshold']) & \
                    (g <= threshold))] \
                    = 1

            self.flags[v]['gradient_depthconditional'] = flag

        if 'spike' in cfg:
            threshold = cfg['spike']
            s = spike(self.input[v])

            if self.saveauxiliary:
                self.auxiliary[v]['spike'] = s

            flag = np.zeros(s.shape, dtype='i1')
            # Flag as 9 any masked input value
            flag[ma.getmaskarray(self.input[v])] = 9
            flag[np.nonzero(s > threshold)] = 4
            flag[np.nonzero(s <= threshold)] = 1
            self.flags[v]['spike'] = flag

        if 'spike_depthconditional' in cfg:
            cfg_tmp = cfg['spike_depthconditional']
            s = spike(self.input[v])
            flag = np.zeros(s.shape, dtype='i1')
            # Flag as 9 any masked input value
            flag[ma.getmaskarray(self.input[v])] = 9
            # ---- Shallow zone -----------------
            threshold = cfg_tmp['shallow_max']
            flag[np.nonzero( \
                    (self['pressure'] <= cfg_tmp['pressure_threshold']) & \
                    (g > threshold))] \
                    = 4
            flag[np.nonzero( \
                    (self['pressure'] <= cfg_tmp['pressure_threshold']) & \
                    (g <= threshold))] \
                    = 1
            # ---- Deep zone --------------------
            threshold = cfg_tmp['deep_max']
            flag[np.nonzero( \
                    (self['pressure'] > cfg_tmp['pressure_threshold']) & \
                    (g > threshold))] \
                    = 4
            flag[np.nonzero( \
                    (self['pressure'] > cfg_tmp['pressure_threshold']) & \
                    (g <= threshold))] \
                    = 1

            self.flags[v]['spike_depthconditional'] = flag

        if 'tukey53H_norm' in cfg:
            """

                I slightly modified the Goring & Nikora 2002. It is
                  expected that CTD profiles has a typical depth
                  structure, with a range between surface and bottom.
            """
            k = cfg['tukey53H_norm']['k']
            s = tukey53H_norm(self.input[v], k)

            if self.saveauxiliary:
                self.auxiliary[v]['tukey53H_norm'] = s

            flag = np.zeros(s.shape, dtype='i1')
            # Flag as 9 any masked input value
            flag[ma.getmaskarray(self.input[v])] = 9
            flag[np.nonzero(s > threshold)] = 4
            flag[np.nonzero(s <= threshold)] = 1
            self.flags[v]['tukey53H_norm'] = flag

        #if 'spike_depthsmooth' in cfg:
        #    from maud.window_func import _weight_hann as wfunc
        #    cfg_tmp = cfg['spike_depthsmooth']
        #    cfg_tmp['dzwindow'] = 10
        #    smooth = ma.masked_all(self.input[v].shape)
        #    z = ped['pressure']
        #    for i in range(len(self.input[v])):
        #        ind = np.nonzero(ma.absolute(z-z[i]) < \
        #                cfg_tmp['dzwindow'])[0]
        #        ind = ind[ind != i]
        #        w = wfunc(z[ind]-z[i], cfg_tmp['dzwindow'])
        #        smooth[i] = (T[ind]*w).sum()/w.sum()

        if 'digit_roll_over' in cfg:
            threshold = cfg['digit_roll_over']
            s = step(self.input[v])

            if self.saveauxiliary:
                self.auxiliary[v]['step'] = s

            flag = np.zeros(s.shape, dtype='i1')
            # Flag as 9 any masked input value
            flag[ma.getmaskarray(self.input[v])] = 9

            flag[np.nonzero(ma.absolute(s) > threshold)] = 4
            flag[np.nonzero(ma.absolute(s) <= threshold)] = 1

            self.flags[v]['digit_roll_over'] = flag

        if 'bin_spike' in cfg:
            bin = bin_spike(self.input[v], cfg['bin_spike'])

            if self.saveauxiliary:
                self.auxiliary[v]['bin_spike'] = bin

        if 'density_inversion' in cfg:
            threshold = cfg['density_inversion']
            ds = densitystep(self['salinity'], self['temperature'],
                    self['pressure'])

            if self.saveauxiliary:
                self.auxiliary[v]['density_step'] = ds

            flag = np.zeros(s.shape, dtype='i1')
            # Flag as 9 any masked input value
            flag[ma.getmaskarray(self.input[v])] = 9

            flag[np.nonzero(ds < threshold)] = 3 # I'm not sure to use 3 or 4.
            flag[np.nonzero(ds >= threshold)] = 1

            self.flags[v]['density_inversion'] = flag

        if 'woa_comparison' in cfg:
            woa = woa_profile(v,
                    self.attributes['datetime'],
                    self.attributes['latitude'],
                    self.attributes['longitude'],
                    self.input['pressure'],
                    cfg['woa_comparison'])

            if woa is None:
                self.logger.warn("%s - WOA is not available at this site" %
                        self.name)
                return

            woa_bias = ma.absolute(self.input[v] - woa['woa_an'])

            if self.saveauxiliary:
                for k in woa.keys():
                    self.auxiliary[v][k] = woa[k]
                self.auxiliary[v]['woa_bias'] = woa_bias
                self.auxiliary[v]['woa_relbias'] = woa_bias/woa['woa_sd']

            self.flags[v]['woa_comparison'] = np.zeros(self.input[v].shape,
                    dtype='i1')
            # Flag as 9 any masked input value
            self.flags[v]['woa_comparison'][ma.getmaskarray(self.input[v])] = 9

            ind = woa_bias/woa['woa_sd'] <= \
                    cfg['woa_comparison']['sigma_threshold']
            self.flags[v]['woa_comparison'][np.nonzero(ind)] = 1
            ind = woa_bias/woa['woa_sd'] > \
                    cfg['woa_comparison']['sigma_threshold']
            self.flags[v]['woa_comparison'][np.nonzero(ind)] = 3

        if 'pstep' in cfg:
            ind = np.isfinite(self.input[v])
            if self.saveauxiliary:
                self.auxiliary[v]['pstep'] = ma.concatenate(
                        [ma.masked_all(1),
                            np.diff(self.input['pressure'][ind])])

    def build_auxiliary(self):
        if not hasattr(self, 'auxiliary'):
            self.auxiliary = {}

        self.auxiliary['common'] = {}
        self.auxiliary['common']['descentPrate'] = \
            descentPrate(self['timeS'], self['pressure'])


class fProfileQC(ProfileQC):
    """ Apply ProfileQC straight from a file.
    """
    def __init__(self, inputfile, cfg=None, saveauxiliary=False, verbose=True,
            logger=None):
        """
        """
        self.logger = logger or logging.getLogger(__name__)
        self.name = 'fProfileQC'

        try:
            # Not the best way, but will work for now. I should pass
            #   the reference for the logger being used.
            input = cnv.fCNV(inputfile, logger=None)
        except CNVError as e:
            #self.attributes['filename'] = basename(inputfile)
            self.logger.error(e.msg)
            raise

        super(fProfileQC, self).__init__(input, cfg=cfg,
                saveauxiliary=saveauxiliary, verbose=verbose)


class ProfileQCed(ProfileQC):
    """
    """
    def __init__(self, input, cfg=None):
        """
        """
        self.name = 'ProfileQCed'
        super(ProfileQCed, self).__init__(input, cfg)

    def keys(self):
        """ Return the available keys in self.data
        """
        return self.input.keys()

    def __getitem__(self, key):
        """ Return the key array from self.data
        """
        if key not in self.flags.keys():
            return self.input[key]
        else:
            f = combined_flag(self.flags[key])
            return ma.masked_array(self.input[key].data, mask=(f!=1))

        raise KeyError('%s not found' % key)



import os
from os.path import expanduser
import re

import numpy as np
from numpy import ma

try:
    import netCDF4
except:
    print("netCDF4 is not available")

try:
    from pydap.client import open_url
    import pydap.lib
    pydap.lib.CACHE = expanduser('~/.cotederc/pydap_cache')
except:
    print("PyDAP is not available")

from scipy.interpolate import RectBivariateSpline, interp1d

# ============================================================================
def woa_profile(var, d, lat, lon, depth, cfg):
    try:
        woa = woa_profile_from_file(var,
                d, lat, lon, depth, cfg)
    except:
        try:
            woa = woa_profile_from_dap(var,
                d, lat, lon, depth, cfg)
        except:
            print "Couldn't make woa_comparison of %s" % var
            return

    return woa


def woa_profile_from_dap(var, d, lat, lon, depth, cfg):
    """
    Monthly Climatologic Mean and Standard Deviation from WOA,
    used either for temperature or salinity.

    INPUTS
        time: [day of the year]
        lat: [-90<lat<90]
        lon: [-180<lon<180]
        depth: [meters]

    Reads the WOA Monthly Climatology NetCDF file and
    returns the corresponding WOA values of salinity or temperature mean and
    standard deviation for the given time, lat, lon, depth.
    """
    if lon<0: lon = lon+360

    url = cfg['url']

    doy = int(d.strftime('%j'))
    dataset = open_url(url)

    dn = (np.abs(doy-dataset['time'][:])).argmin()
    xn = (np.abs(lon-dataset['lon'][:])).argmin()
    yn = (np.abs(lat-dataset['lat'][:])).argmin()

    if re.match("temperature\d?$", var):
        mn = ma.masked_values(dataset.t_mn.t_mn[dn,:,yn,xn].reshape(dataset['depth'].shape[0]), dataset.t_mn.attributes['_FillValue'])
        sd = ma.masked_values(dataset.t_sd.t_sd[dn,:,yn,xn].reshape(dataset['depth'].shape[0]), dataset.t_sd.attributes['_FillValue'])
        #se = ma.masked_values(dataset.t_se.t_se[dn,:,yn,xn].reshape(dataset['depth'].shape[0]), dataset.t_se.attributes['_FillValue'])
        # Use this in the future. A minimum # of samples
        dd = ma.masked_values(dataset.t_dd.t_dd[dn,:,yn,xn].reshape(dataset['depth'].shape[0]), dataset.t_dd.attributes['_FillValue'])
    elif re.match("salinity\d?$", var):
        mn = ma.masked_values(dataset.s_mn.s_mn[dn,:,yn,xn].reshape(dataset['depth'].shape[0]), dataset.s_mn.attributes['_FillValue'])
        sd = ma.masked_values(dataset.s_sd.s_sd[dn,:,yn,xn].reshape(dataset['depth'].shape[0]), dataset.s_sd.attributes['_FillValue'])
        dd = ma.masked_values(dataset.s_dd.s_dd[dn,:,yn,xn].reshape(dataset['depth'].shape[0]), dataset.s_dd.attributes['_FillValue'])
    zwoa = ma.array(dataset.depth[:])

    ind = (depth<=zwoa.max()) & (depth>=zwoa.min())
    # Mean value profile
    f = interp1d(zwoa[~ma.getmaskarray(mn)].compressed(), mn.compressed())
    mn_interp = ma.masked_all(depth.shape)
    mn_interp[ind] = f(depth[ind])
    # The stdev profile
    f = interp1d(zwoa[~ma.getmaskarray(sd)].compressed(), sd.compressed())
    sd_interp = ma.masked_all(depth.shape)
    sd_interp[ind] = f(depth[ind])

    output = {'woa_an': mn_interp, 'woa_sd': sd_interp}

    return output

# ============================================================================
def woa_profile_from_file(var, d, lat, lon, depth, cfg):
    """
    Monthly Climatologic Mean and Standard Deviation from WOA,
    used either for temperature or salinity.

    INPUTS
        time: [day of the year]
        lat: [-90<lat<90]
        lon: [-180<lon<180]
        depth: [meters]

    Reads the WOA Monthly Climatology NetCDF file and
    returns the corresponding WOA values of salinity or temperature mean and
    standard deviation for the given time, lat, lon, depth.
    """
    if lon<0: lon = lon+360

    doy = int(d.strftime('%j'))
    nc = netCDF4.Dataset(expanduser(cfg['file']), 'r')

    # Get the nearest point. In the future interpolate.
    dn = (np.abs(doy - nc.variables['time'][:])).argmin()
    xn = (np.abs(lon - nc.variables['lon'][:])).argmin()
    yn = (np.abs(lat - nc.variables['lat'][:])).argmin()

    vars = cfg['vars']

    climdata = {}
    for v in vars:
        climdata[v] = ma.masked_values(nc.variables[vars[v]][dn, :, yn, xn], nc.variables[vars[v]]._FillValue)

    zwoa = ma.array(nc.variables['depth'][:])

    ind_z = (depth<=zwoa.max()) & (depth>=zwoa.min())
    output = {}
    # Mean value profile
    for v in vars:
        # interp1d can't handle masked values
        ind_valid = ~ma.getmaskarray(climdata[v])
        f = interp1d(zwoa[ind_valid], climdata[v][ind_valid])
        output[v] = ma.masked_all(depth.shape)
        output[v][ind_z] = f(depth[ind_z])
    ## The stdev profile
    #f = interp1d(zwoa[~ma.getmaskarray(sd)].compressed(), sd.compressed())
    #sd_interp = ma.masked_all(depth.shape)
    #sd_interp[ind] = f(depth[ind])

    return output



# -*- coding: utf-8 -*-

"""
"""

import numpy as np
from numpy import ma


def tukey53H(x):
    """Spike test Tukey 53H from Goring & Nikora 2002
    """
    N = len(x)

    u1 = ma.masked_all(N)
    for n in range(N-4):
        if x[n:n+5].any():
            u1[n+2] = ma.median(x[n:n+5])

    u2 = ma.masked_all(N)
    for n in range(N-2):
        if u1[n:n+3].any():
            u2[n+1] = ma.median(u1[n:n+3])

    u3 = ma.masked_all(N)
    u3[1:-1] = 0.25*(u2[:-2] + 2*u2[1:-1] + u2[2:])

    Delta = ma.absolute(x-u3)

    return Delta


def tukey53H_norm(x, k=1.5, l=12):
    """Spike test Tukey53H() normalized by the std of the low pass

       l is the number of observations. The default l=12 is trully not
         a big number, but this test foccus on spikes, therefore, any
         variability longer than 12 is something else.
    """
    Delta = tukey53H(x)

    w = np.hamming(l)
    sigma = (np.convolve(x, w, mode='same') / w.sum()).std()

    return Delta/(k*sigma)



# -*- coding: utf-8 -*-

"""
"""

import numpy as np
from numpy import ma
#from scipy.stats import norm, rayleigh, expon, halfnorm, exponpow, exponweib
from scipy.stats import exponweib
#from scipy.stats import kstest

from cotede.utils import ProfilesQCPandasCollection
from cotede.misc import combined_flag
from cotede.humanqc import HumanQC

def fit_tests(features, qctests, ind=True, q=0.90, verbose=False):
    """

        Input:
          features: a dictionary like with the numerical results from the
              QC tests. For example, the gradient test values, not the
              flags, but the floats itself, like
              {'gradient': ma.array([.23, .12, .08]), 'spike': ...}
          qctests: The name of the tests to fit. They must be in
              features.keys().
              ['gradient', 'spike', 'woa_bias']
          ind: The features values positions to be considered in the fit.
              It's usefull to eliminate out of range data, or to
              restrict to a subset of the data, like in the calibration
              procedure.
          q: The lowest percentile to be considered. For example, .90
              means that only the top 10% data (i.e. percentiles higher
              than .90) are considered in the fitting.
    """
    output = {}
    for test in qctests:
        samp = features[test][ind & np.isfinite(features[test])]
        ind_top = samp > samp.quantile(q)
        param = exponweib.fit(samp[ind_top])
        output[test] = {'param':param,
                'qlimit': samp.quantile(q)}

        if verbose == True:
            import pylab
            x = np.linspace(samp[ind_top].min(), samp[ind_top].max(), 100)
            pdf_fitted = exponweib.pdf(x, *param[:-2], loc=param[-2], scale=param[-1])
            pylab.plot(x,pdf_fitted,'b-')
            pylab.hist(ma.array(samp[ind_top]), 100, normed=1, alpha=.3)
            pylab.title(test)
            pylab.show()

    return output


def estimate_anomaly(aux, params):
    prob = ma.ones(aux.shape[0])
    for t in params.keys():
        param = params[t]['param']
        ind = np.array(np.isfinite(aux[t]))
        prob[ind] = prob[ind] * \
                exponweib.sf(aux[t][ind], *param[:-2], loc=param[-2], scale=param[-1])
    return prob


def estimate_p_optimal(prob, qc, verbose=False):
    err = []
    P = 10.**np.arange(-12, 0, 0.1)
    for p in P:
        false_negative = (prob < p) & (qc == True)
        false_positive = (prob > p) & (qc == False)
        err.append(np.nonzero(false_negative)[0].size + \
                np.nonzero(false_positive)[0].size)
    err = np.array(err)
    if verbose == True:
        pylab.plot(P, err , 'b'); pylab.show()
    return P[err.argmin()], float(err.min())/prob.size#, {'P': P, 'err': err}


def adjust_anomaly_coefficients(flag_ref, qctests, aux, q=0.90, verbose=False):
    """ Adjust coeficients for Anomaly Detection, and estimate error

        Inputs:
            flag_ref: Reference index. What the Anomaly Detection will try
                   to reproduce. Uses the True and Falses from flag_ref
                   to partition the data to be used to fit, to adjust
                   and to estimate the error.
            qctests: The tests used by the Anomaly Detection. One curve will
                   be fit for each test.
            aux: The auxiliary tests results from the ProfileQCCollection. It
                   is expected that the qctests are present in aux.
            q: The top q extreme tests results to be used on Anom. Detect.
                 For example q=0 will use all the data, while q=0.9 (default)
                 will use the percentile of 0.9, i.e. the top 10% values.

            Output: Returns a dictionary with
                err:
                err_ratio:
                false_negative:
                false_positive:
                p_optimal:
                params:

            Use the functions:
                split_data_groups()
                fit_tests()
                estimate_anomaly()
                estimate_p_optimal()

    """
    indices = split_data_groups(flag_ref)
    params = fit_tests(aux, qctests, indices['ind_fit'], q=q,
            verbose=verbose)
    prob = estimate_anomaly(aux, params)
    if verbose == True:
        pylab.hist(prob); pylab.show()

    p_optimal, test_err = estimate_p_optimal(prob[indices['ind_test']],
            flag_ref[indices['ind_test']])

    # I can extract only .data, since split_data_groups already eliminated
    #   all non valid positions.
    false_negative = (prob[indices['ind_err']] < p_optimal) & \
        (flag_ref[indices['ind_err']].data == True)
    false_positive = (prob[indices['ind_err']] > p_optimal) & \
        (flag_ref[indices['ind_err']].data == False)
    err = np.nonzero(false_negative)[0].size + \
            np.nonzero(false_positive)[0].size
    err_ratio = float(err)/prob[indices['ind_err']].size
    false_negative = (prob < p_optimal) & \
        (flag_ref.data == True) & (ma.getmaskarray(flag_ref)==False)
    false_positive = (prob > p_optimal) & \
        (flag_ref.data == False) & (ma.getmaskarray(flag_ref)==False)

    output = {'false_negative': false_negative,
            'false_positive': false_positive,
            'prob': prob,
            'p_optimal': p_optimal,
            'err': err,
            'err_ratio': err_ratio,
            'params': params}

    return output


def split_data_groups(ind):
    """ Splits randomly the indices into fit, test and error groups

        Return 3 indices set:
            - ind_fit with 60% of the good
            - ind_test with 20% of the good and 50% of the bad
            - ind_eval with 20% of the good and 50% of the bad
    """
    assert ind.dtype == 'bool'

    N = ind.size
    ind_base = np.zeros(N) == 1
    ind_valid = ~ma.getmaskarray(ind)

    # ==== Good data ==================
    ind_good = np.nonzero((ind == True) & ind_valid)[0]
    N_good = ind_good.size
    perm = np.random.permutation(N_good)
    N_fit = int(round(N_good*.6))
    N_test = int(round(N_good*.2))
    ind_fit = ind_base.copy()
    ind_fit[ind_good[perm[:N_fit]]] = True
    ind_test = ind_base.copy()
    ind_test[ind_good[perm[N_fit:-N_test]]] = True
    ind_err = ind_base.copy()
    ind_err[ind_good[perm[-N_test:]]] = True

    # ==== Bad data ===================
    ind_bad = np.nonzero((ind == False) & ind_valid)[0]
    N_bad = ind_bad.size
    perm = np.random.permutation(N_bad)
    N_test = int(round(N_bad*.5))
    ind_test[ind_bad[perm[:N_test]]] = True
    ind_err[ind_bad[perm[N_test:]]] = True

    output = {'ind_fit': ind_fit, 'ind_test': ind_test, 'ind_err': ind_err}
    return output


def flags2binflag(flags, reference_flags):
    """
    """

    # The different flags must have same ammount of data.
    N = len(flags[flags.keys()[0]])
    for f in flags:
        assert len(flags[f]) == N

    flags = combined_flag(flags, reference_flags)

    output = ma.masked_all(N, dtype='bool')
    output[flags==1] = True
    output[(flags==3) | (flags==4)] = False

    return output


def order_unprobable_profiles(datadir, varname):
    """ Return a list of profiles ordered by less probable data
    """
    qctests = ['gradient', 'step', 'tukey53H_norm', 'woa_relbias']
    reference_flags = ['global_range', 'gradient_depthconditional',
            'spike_depthconditional', 'digit_roll_over']
    #hardlimit_flags = ['global_range']

    db = ProfilesQCPandasCollection(datadir, saveauxiliary=True)
    aux = db.auxiliary[varname]

    params = fit_tests(aux, qctests, q=.9,)
    prob = estimate_anomaly(aux, params)

    profilename = np.asanyarray(db.data['profilename'])
    output = []
    for pid in np.unique(profilename):
        output.append([pid, min(prob[profilename == pid])])

    return [x[0] for x in sorted(output, key=lambda x: x[1])]


def calibrate_anomaly_detection(datadir, varname):
    """
    """
    qctests = ['gradient', 'step', 'tukey53H_norm', 'woa_relbias']
    reference_flags = ['global_range', 'gradient_depthconditional',
            'spike_depthconditional', 'digit_roll_over']
    #hardlimit_flags = ['global_range']

    db = ProfilesQCPandasCollection(datadir, saveauxiliary=True)
    aux = db.auxiliary[varname]

    binflag = flags2binflag(db.flags[varname], reference_flags)
    # Remove the value out of the possible range.
    ind_outofrange = np.nonzero(db.flags[varname]['global_range'] != 1)
    binflag.mask[ind_outofrange] = True

    result = adjust_anomaly_coefficients(binflag, qctests, aux)

    return result


def human_calibrate_mistakes(datadir, varname, niter=5):
    """
    """
    qctests = ['gradient', 'step', 'tukey53H_norm', 'woa_relbias']
    reference_flags = ['global_range', 'gradient_depthconditional',
            'spike_depthconditional', 'digit_roll_over']
    #hardlimit_flags = ['global_range']

    db = ProfilesQCPandasCollection(datadir, saveauxiliary=True)

    # Remove the value out of the possible range.
    ind_valid = db.flags[varname]['global_range'] == 1
    data = db.data.loc[ind_valid, ['profileid', 'pressure', varname]]
    flags = db.flags[varname][ind_valid]
    aux = db.auxiliary[varname][ind_valid]

    binflag = flags2binflag(flags, reference_flags)

    result = adjust_anomaly_coefficients(binflag, qctests, aux)
    error_log = [{'err': result['err'], 'err_ratio': result['err_ratio'],
                     'p_optimal': result['p_optimal']}]

    # I don't like this approach. Improve this in the future.
    ind_humanqc = binflag.copy()
    for i in range(niter):
        # Profiles with any failure
        mistakes = (result['false_positive'] | result['false_negative'])
        profileids = data['profileid'].iloc[mistakes].unique()
        if len(profileids) == 0:
            break
        # Only 5 profiles each time
        for pid in np.random.permutation(profileids)[:5]:
            print("Profile: %s" % pid)
            ind_p = np.array(data.profileid == pid)
            print data.profilename[ind_p].iloc[0]
            mistakes = (result['false_positive'][ind_p] |
                    result['false_negative'][ind_p])
            h = HumanQC().eval(
                    np.array(data[varname][ind_p]),
                    np.array(data['pressure'][ind_p]),
                    baseflag=ind_humanqc[ind_p],
                    fails=mistakes)#, doubt = ind_doubt[ind])
            ind_humanqc[np.nonzero(ind_p)[0][h=='good']] = True
            ind_humanqc[np.nonzero(ind_p)[0][h=='bad']] = False
            ind_humanqc.mask[np.nonzero(ind_p)[0][h=='doubt']] = True

            result = adjust_anomaly_coefficients(ind_humanqc, qctests, aux)
            error_log.append({'err': result['err'],
                'err_ratio': result['err_ratio'],
                'p_optimal': result['p_optimal']})

        print error_log[-2]
        print error_log[-1]
    return {'ind_humanqc':ind_humanqc, 'error_log': error_log,
            'result': result}



# -*- coding: utf-8 -*-

"""

#http://www.ngdc.noaa.gov/mgg/global/relief/ETOPO1/data/bedrock/grid_registered/netcdf/ETOPO1_Bed_g_gmt4.grd.gz

supportdata.download_file('http://opendap.ccst.inpe.br/Climatologies/ETOPO/etopo5.cdf','309bef6916aee6e12563d3f8c1f27503')
"""

import urllib2
import hashlib
import os
import shutil
from tempfile import NamedTemporaryFile

def download_file(url, md5hash):
    """ Download data file from web

        IMPROVE it to automatically extract gz files
    """
    download_block_size = 2 ** 16

    assert type(md5hash) is str

    d = os.path.expanduser("~/.cotederc/data")
    if not os.path.exists(d):
        os.makedirs(d)

    hash = hashlib.md5()

    fname = os.path.join(d, os.path.basename(url))
    if os.path.isfile(fname):
        h = hashlib.md5(open(fname, 'rb').read()).hexdigest()
        if h == md5hash:
            print("Was previously downloaded: %s" % fname)
            return
        else:
            assert False, "%s already exist but doesn't match the hash: %s" % \
                    (fname, md5hash)

    remote = urllib2.urlopen(url)

    with NamedTemporaryFile(delete=False) as f:
        try:
            bytes_read = 0
            block = remote.read(download_block_size)
            while block:
                f.write(block)
                hash.update(block)
                bytes_read += len(block)
                block = remote.read(download_block_size)
        except:
            if os.path.exists(f.name):
                os.remove(f.name)
                raise

    h = hash.hexdigest()
    if h != md5hash:
        os.remove(f.name)
        print("Downloaded file doesn't match. %s" % h)
        assert False, "Downloaded file (%s) doesn't match with expected hash (%s)" % \
                (fname, md5hash)

    shutil.move(f.name, fname)
    print("Downloaded: %s" % fname)

def download_supportdata():
    print("This can take several minutes, depending on the network bandwidth. Sorry, in the future I'll include a progress bar.")
    download_file('http://opendap.ccst.inpe.br/Climatologies/ETOPO/etopo5.cdf','309bef6916aee6e12563d3f8c1f27503')
    download_file('http://data.nodc.noaa.gov/thredds/fileServer/woa/WOA09/NetCDFdata/temperature_seasonal_5deg.nc','271f66e8dea4dfef7db99f5f411af330')
    download_file('http://data.nodc.noaa.gov/thredds/fileServer/woa/WOA09/NetCDFdata/salinity_seasonal_5deg.nc','1d2d1982338c688bdd18069d030ec05f')

def download_testdata(filename):

    #d = os.path.expanduser("~/.cotederc/testdata")
    d = os.path.expanduser("~/.cotederc/data")
    if not os.path.exists(d):
        os.makedirs(d)

    test_files = {
            'dPIRX010.cnv': {
                "url": "https://dl.dropboxusercontent.com/u/26063625/seabird/dPIRX010.cnv",
                "md5": "8691409accb534c83c8bd412afbdd285"},
            'dPIRX003.cnv': {
                "url": "https://dl.dropboxusercontent.com/u/26063625/seabird/dPIRX003.cnv",
                "md5": "4b941b902a3aea7d99e1cf4c78c51877"},
            'PIRA001.cnv': {
                "url": "https://dl.dropboxusercontent.com/u/26063625/seabird/PIRA001.cnv",
                "md5": "5ded777144300b63c8775b1d7f033f92"},
            'TSG_PIR_001.cnv': {
                "url": "https://dl.dropboxusercontent.com/u/26063625/seabird/TSG_PIR_001.cnv",
                "md5": "2950ccb9f77e0802557b011c63d2e39b"},
            }

    assert filename in test_files.keys(), \
            "%s is not a valid test file" % filename

    download_file(test_files[filename]["url"], test_files[filename]["md5"])
    datafile = os.path.join(os.path.expanduser("~/.cotederc/data/"),
            filename)

    return datafile



# -*- coding: utf-8 -*-

from cotede.utils import get_depth

def location_at_sea(data, cfg):
    """ Evaluate if location is at sea.

        Interpolate the depth from ETOPO for the data position.
          If local "height" is negative, means lower than sea
          level, hence at sea.
    """
    assert hasattr(data, 'attributes'), "Missing attributes"
    assert ('latitude' in data.attributes), \
            "Missing latitude in attributes"
    assert ('longitude' in data.attributes), \
            "Missing longitude in attributes"

    depth = get_depth(data.attributes['latitude'],
            data.attributes['longitude'],
            cfg=cfg)

    if depth < 0:
        return 1
    elif flag >= 0:
        return 3
    else:
        return 0



# -*- coding: utf-8 -*-

""" The tests itself to QC the data
"""

import numpy as np
from numpy import ma

def step(x):
    y = ma.masked_all(x.shape, dtype=x.dtype)
    y[1:] = ma.diff(x)
    return y


def gradient(x):
    y = ma.masked_all(x.shape, dtype=x.dtype)
    y[1:-1] = np.abs(x[1:-1] - (x[:-2] + x[2:])/2.0)
    # ATENTION, temporary solution
    #y[0]=0; y[-1]=0
    return y


def spike(x):
    y = ma.masked_all(x.shape, dtype=x.dtype)
    y[1:-1] = np.abs(x[1:-1] - (x[:-2] + x[2:])/2.0) - \
                np.abs((x[2:] - x[:-2])/2.0)
    # ATENTION, temporary solution
    #y[0]=0; y[-1]=0
    return y


def bin_spike(x, l):
    """

        Dummy way to avoid warnings when x[ini:fin] are all masked.
        Improve this in the future.
    """
    N = len(x)
    bin = ma.masked_all(N)
    half_window = l/2
    for i in range(half_window, N-half_window):
        ini = max(0, i - half_window)
        fin = min(N, i + half_window)
        if ~x[ini:fin].mask.any():
            bin[i] = x[i] - ma.median(x[ini:fin])
            #bin_std[i] = (T[ini:fin]).std()

    return bin


def densitystep(S, T, P):
    """
    """
    assert S.shape == T.shape
    assert S.shape == P.shape
    try:
        import gsw
        rho0 = gsw.pot_rho_t_exact(S, T, P, 0)
        ds = ma.masked_all(S.shape, dtype=S.dtype)
        ds[1:] = np.sign(np.diff(P))*np.diff(rho0)
        return ds

    except ImportError:
        print("Package gsw is required and is not available.")



#!/usr/bin/env python

from cotede.utils.utils import *
from cotede.utils.woa import *
from cotede.utils.profilescollection import ProfilesQCCollection
from cotede.utils.profilescollection import ProfilesQCPandasCollection

import logging



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
#from distutils.core import setup

import os
import sys

# ============================================================================
from setuptools.command.test import test as TestCommand
class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True
    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)
# ============================================================================

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()
NEWS = open(os.path.join(here, 'NEWS.txt')).read()

install_requires = [
    'numpy>=1.1',
    'seabird>=0.5.8',
    'scipy',
    ]

version = '0.12.5'

setup(
    name='cotede',
    version=version,
    author='Guilherme Castelão',
    author_email='guilherme@castelao.net',
    packages=['cotede', 'cotede.qctests', 'cotede.utils', 'cotede.humanqc'],
    url='http://cotede.castelao.net',
    license='License :: OSI Approved :: BSD License',
    description='Quality Control of CTD profiles',
    long_description=open('README.rst').read(),
    install_requires=install_requires,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 2',
        ],
    keywords='CTD SeaBird QualityControl oceanography hydrography',
    #package_dir = {'': './'},
    include_package_data=True,
    zip_safe=False,
    platforms=['any'],
    scripts=["bin/ctdqc"],
    tests_require=['pytest'],
    cmdclass={'test': PyTest},
)



import os
from os.path import expanduser
import re

import numpy as np
from numpy import ma

try:
    import netCDF4
except:
    print("netCDF4 is not available")

try:
    from pydap.client import open_url
    import pydap.lib
    pydap.lib.CACHE = '.cache'
except:
    print("PyDAP is not available")

from scipy.interpolate import RectBivariateSpline, interp1d


def make_file_list(inputdir, inputpattern):
    """ Search inputdir recursively for inputpattern
    """
    inputfiles = []
    for dirpath, dirnames, filenames in os.walk(inputdir):
        for filename in filenames:
            if re.match(inputpattern, filename):
                inputfiles.append(os.path.join(dirpath, filename))
    inputfiles.sort()
    return inputfiles


def get_depth(lat, lon, cfg):
    """

    ATTENTION, conceptual error on the data near by Greenwich.
    url='http://opendap.ccst.inpe.br/Climatologies/ETOPO/etopo5.cdf'

    If I ever need to get depth from multiple points, check the history
      of this file. One day it was like that.
    """
    assert type(lat) in [int, float]
    assert type(lon) in [int, float]

    # if lat.shape != lon.shape:
    #            print "lat and lon must have the same size"

    try:
        try:
            etopo = netCDF4.Dataset(expanduser(cfg['file']))
        except:
            etopo = netCDF4.Dataset(expanduser(cfg['url']))
        x = etopo.variables['ETOPO05_X'][:]
        y = etopo.variables['ETOPO05_Y'][:]
    except:
        etopo = open_url(cfg['url']).ROSE
        x = etopo.ETOPO05_X[:]
        y = etopo.ETOPO05_Y[:]

    if lon < 0:
        lon += 360

    iini = (abs(lon - x)).argmin() - 2
    ifin = (abs(lon - x)).argmin() + 2
    jini = (abs(lat - y)).argmin() - 2
    jfin = (abs(lat - y)).argmin() + 2

    assert (iini >= 0) or (iini <= len(x)) or \
        (jini >= 0) or (jini <= len(y)), \
        "Sorry not ready to handle too close to boundaries"

    try:
        z = etopo.variables['ROSE'][jini:jfin, iini:ifin]
    except:
        z = etopo.ROSE[jini:jfin, iini:ifin]

    interpolator = RectBivariateSpline(x[iini:ifin], y[jini:jfin], z.T)
    return interpolator(lon, lat)[0][0]


# ============================================================================
def savePQCCollection_pandas(db, filename):
    """ Save

        To Do:
            - Save the files in a tmp file
            - As it saves, creates a md5 of each file
            - Put everything together in a tar.bz2, including the md5 list
            - Delete the tmp file
    """
    import os
    import tempfile
    import tarfile
    import shutil
    import hashlib
    # tar = tarfile.open("%s.tar.bz2" % filename, "w:bz2")
    tar = tarfile.open(filename, "w:bz2")
    tmpdir = tempfile.mkdtemp()

    try:
        # Data
        f = "%s/data.hdf" % (tmpdir)
        db.data.to_hdf(f, 'df')
        tar.add(f, arcname='data.hdf')
        # hashlib.md5(open(f, 'rb').read()).digest()
        # hashlib.sha256(open(f, 'rb').read()).digest()
        # Flags
        p = os.path.join(tmpdir, 'flags')
        os.mkdir(p)
        for k in db.flags.keys():
            f = os.path.join(p, "flags_%s.hdf" % k)
            db.flags[k].to_hdf(f, 'df')
            tar.add(f, arcname="flags/flags_%s.hdf" % k)
        if hasattr(db, 'auxiliary'):
            p = os.path.join(tmpdir, 'aux')
            os.mkdir(p)
            for k in db.auxiliary.keys():
                f = os.path.join(p, "aux_%s.hdf" % k)
                db.auxiliary[k].to_hdf(f, 'df')
                tar.add(f, arcname="aux/aux_%s.hdf" % k)
        tar.close()
    except:
        shutil.rmtree(tmpdir)
        raise
        print "Problems saving the data"
        shutil.rmtree("%s.tar.bz2" % filename)
    finally:
        shutil.rmtree(tmpdir)
